/*  
 *  snake3d, a snake game in a 3d world.
 *  Copyright (C) 2005-2006 Yair Yarom
 *  
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or (at your option) any later
 *  version.
 *  
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License along with
 *  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 *  Place, Suite 330, Boston, MA 02111-1307 USA
 *  
 *  http://sourceforge.net/projects/worms3d
 *  
 */

#if (defined __APPLE__ && defined __MACH__)
#include <SDL_net/SDL_net.h>
#else
#include <SDL_net.h>
#endif
#include <string.h>

#include "client.hpp"
#include "network.hpp"
#include "global.hpp"
#include "preferences.hpp"
#include "packets.hpp"
#include "menu.hpp"
#include "board.hpp"
#include "snake.hpp"

static SDLNet_SocketSet socketSet = NULL;
static TCPsocket connection = NULL;

void menuClient();
void gameClient();

int client(void* data) {
    IPaddress ip;
    ip.host = prefs.clientConnection.host;
    ip.port = prefs.clientConnection.port;

    connection = SDLNet_TCP_Open(&ip);
    if (!connection) {
	fprintf(stderr, "Connection error: %s\n", SDLNet_GetError());
	networkError = true;
	return 1;
    }

    socketSet = SDLNet_AllocSocketSet(1);
    if (!socketSet) {
	fprintf(stderr, "Connection error: %s\n", SDLNet_GetError());
	networkError = true;
	SDLNet_TCP_Close(connection);
	return 2;
    }
    
    SDLNet_TCP_AddSocket(socketSet, connection);
    for (int i = 0; i < MAX_SNAKES; i++)
	prefs.ai[i] = ai_remote;
    

    
    while (!networkError) {
	switch (p_mode) {
	  case pm_menu: menuClient(); break;
	  case pm_game: gameClient(); break;
	  case pm_quit: networkError = true; break;
	}
    }

    SDLNet_FreeSocketSet(socketSet);
    SDLNet_TCP_Close(connection);
    return 0;
}


void clientSendSnakeConfig(int snake) {
    sendPacket_SnakeConfig(connection, snake);
}

void clientSendStart() {
    sentTick = true;
    sendPacket_Start(connection);
}

void clientSendTick() {
    //printf("client sent tick\n");
    sentTick = true;
    sendPacket_Tick(connection);
}

void clientSendStop() {
    sendPacket_Stop(connection);
}

void clientSendChangeDir(int snake, direction dir) {
    //printf("client sent changed dir\n");
    sendPacket_ChangeDir(connection, snake, dir);
}

void clientSendReset() {
    sendPacket_Reset(connection);
}

void clientSendPause(bool pause) {
    sendPacket_Pause(connection, pause);
}

void menuClient() {
    while (p_mode == pm_menu && !networkError) {
	int active = SDLNet_CheckSockets(socketSet, BASE_INTERVAL * 5);
	if (active < 0) {
	    networkError = true;
	    break;
	}
	if (active > 0) {
	    DEBUG_PRINT("client activity: menu mode");
	    unsigned char packetType;
	    //SDL_mutexP(netLock);
            SDL_SemWait(netLock);
	    int received = SDLNet_TCP_Recv(connection, &packetType, 1);
	    //SDL_mutexV(netLock);
            SDL_SemPost(netLock);
	    if (received < 0) {
		DEBUG_PRINT2(": error - %s\n", SDLNet_GetError());
                perror("Network error");
		networkError = true;
	    } else if (received == 0) {
		DEBUG_PRINT(": closed connection");
		networkError = true;
	    } else {
		switch (packetType) {
		  case ASSIGN_PACKET:
		      DEBUG_PRINT(": assign");
		      unsigned char assign;
		      if (recvPacket_Assign(connection, &assign)) {
			  DEBUG_PRINT2(": snake %i", assign + 1);
			  prefs.ai[assign] = ai_human;
			  strncpy(names[assign], prefs.name, 20);
			  if (!sendPacket_SnakeConfig(connection, assign)) {
			      DEBUG_PRINT(": failed sending snake config");
			      networkError = true;
			  }
		      } else {
			  DEBUG_PRINT2(": failed establishing connection - %s", (assign == 0 ? "receiving" : " version"));
			  networkError = true;
		      }
		      break;

		  case SNAKE_CONFIG_PACKET:
		      DEBUG_PRINT(": snake config");
		      unsigned char snake;
		      recvPacket_SnakeConfig(connection, &snake);
		      DEBUG_PRINT2(": snake %i", snake);
		      break;

		  case BOARD_CONFIG_PACKET:
		      DEBUG_PRINT(": board config");
		      recvPacket_BoardConfig(connection, clientFood);
		      rebuildBoard = true;
		      break;

		  case START_PACKET:
		      DEBUG_PRINT(": start");
		      recvPacket_Start(connection);
		      p_mode = pm_game;
		      break;

		  case STOP_PACKET:
		      DEBUG_PRINT(": stop");
		      recvPacket_Stop(connection);
		      rebuildBoard = true;
		      p_mode = pm_menu;
		      break;

		  case SPEED_PACKET:
		      DEBUG_PRINT(": speed ");
		      unsigned char speed;
		      recvPacket_Speed(connection, &speed);
		      prefs.stepInterval = speed;
		      DEBUG_PRINT2("%i", speed);
		      break;

		  case TOGGLE_FINE_TUNE_PACKET:
		      DEBUG_PRINT(": toggle fine tune");
		      unsigned char fineTune;
		      recvPacket_ToggleFineTune(connection, &fineTune);
		      prefs.fineTune = fineTune;
		      DEBUG_PRINT2("%i", fineTune);
		      break;
		      
		  default:
		      DEBUG_PRINT(": error, unknown packet");
		      networkError = true;
		}
	    }
	    DEBUG_PRINT("\n");
	}
    }
}

void gameClient() {
    while (p_mode == pm_game && !networkError) {
	int active = SDLNet_CheckSockets(socketSet, BASE_INTERVAL * 5);
	if (active < 0) {
	    networkError = true;
	    break;
	}
	if (active > 0) {
	    DEBUG_PRINT("client activity: game mode");
	    unsigned char packetType;
	    //SDL_mutexP(netLock);
            SDL_SemWait(netLock);
	    int received = SDLNet_TCP_Recv(connection, &packetType, 1);
	    //SDL_mutexV(netLock);
            SDL_SemPost(netLock);
	    if (received < 0) {
		DEBUG_PRINT2(": error - %s\n", SDLNet_GetError());
                perror("Network error");
		networkError = true;
	    } else if (received == 0) {
		DEBUG_PRINT(": closed connection");
		networkError = true;
	    } else {
		unsigned char snake;
		switch (packetType) {
		  case TICK_PACKET:
		      DEBUG_PRINT(": tick");
		      recvPacket_Tick(connection);
		      //SDL_mutexP(stepLock);
                      //SDL_SemWait(stepLock);
		      lockStep("got tick");
		      mainBoard->step();
		      stepPercent = 0;
		      unlockStep("end tick");
		      recvTick = true;
		      break;

		  case CHANGE_DIR_PACKET:
		      DEBUG_PRINT(": change direction");
		      direction dir;
		      recvPacket_ChangeDir(connection, &snake, &dir);
		      DEBUG_PRINT2(" snake %i", snake);
		      //while (recvTick);
		      //SDL_mutexP(stepLock);
                      //SDL_SemWait(stepLock);
		      lockStep("got change dir");
		      snakes[snake]->changeDirection(dir);
		      //SDL_mutexV(stepLock);
                      //SDL_SemPost(stepLock);
		      unlockStep("end change dir");
		      break;

		  case INC_PACKET:
		      DEBUG_PRINT(": inc ");
		      recvPacket_Inc(connection, &snake);
		      DEBUG_PRINT2(" snake %i", snake);
		      //while (recvTick);
		      //SDL_mutexP(stepLock);
                      //SDL_SemWait(stepLock);
		      lockStep("got inc");
		      snakes[snake]->inc();
		      //SDL_mutexV(stepLock);
                      //SDL_SemPost(stepLock);
		      unlockStep("end inc");
		      break;

		  case FOOD_PACKET:
		      DEBUG_PRINT(": food");
		      recvPacket_Food(connection, clientFood);
		      mainBoard->replaceFood(clientFood);
		      break;

		  case STOP_PACKET:
		      DEBUG_PRINT(": stop");
		      recvPacket_Stop(connection);
		      rebuildBoard = true;
		      p_mode = pm_menu;
		      break;

		  case SPEED_PACKET:
		      DEBUG_PRINT(": speed ");
		      unsigned char speed;
		      recvPacket_Speed(connection, &speed);
		      recentSpeedChange = true;
		      fineTuneSpeed = 0;
		      prefs.stepInterval = speed;
		      DEBUG_PRINT2("%i", speed);
		      break;

		  case RESET_PACKET:
		      DEBUG_PRINT(": reset");
		      recvPacket_Reset(connection);
		      clientSendReset();
		      //SDL_SemWait(stepLock);
		      lockStep("got reset");
		      mainBoard->resetSnakes();
		      sentTick = false;
		      //SDL_SemPost(stepLock);
		      unlockStep("end reset");
		      break;

		  case FINE_TUNE_PACKET:
		      DEBUG_PRINT(": fine tune ");
		      unsigned char fineTune;
		      recvPacket_FineTune(connection, &fineTune);
		      if (fineTuneSpeed - fineTune < - MAX_FINE_TUNE)
			  fineTuneSpeed = - MAX_FINE_TUNE;
		      else 
			  fineTuneSpeed -= fineTune;
		      if (SPEEDS[prefs.stepInterval] - fineTuneSpeed < 0)
			  fineTuneSpeed = 1 - SPEEDS[prefs.stepInterval];
		      DEBUG_PRINT2("%i", fineTune);
		      break;

		  case PAUSE_PACKET:
		      DEBUG_PRINT(": pause");
		      unsigned char rpause;
		      recvPacket_Pause(connection, &rpause);
		      paused = rpause;
		      break;
		      
		  default:
		      DEBUG_PRINT2(": error, unknown packet %i", packetType);
		      networkError = true;
		}
	    }
	    DEBUG_PRINT("\n");
	}
    }
}

