/*  
 *  snake3d, a snake game in a 3d world.
 *  Copyright (C) 2005-2006 Yair Yarom
 *  
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or (at your option) any later
 *  version.
 *  
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License along with
 *  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 *  Place, Suite 330, Boston, MA 02111-1307 USA
 *  
 *  http://sourceforge.net/projects/worms3d
 *  
 */

#if (defined __APPLE__ && defined __MACH__)
#include <SDL_net/SDL_net.h>
#else
#include <SDL_net.h>
#endif

#include "preferences.hpp"
#include "global.hpp"
#include "server.hpp"
#include "network.hpp"
#include "packets.hpp"
#include "menu.hpp"
#include "snake.hpp"

static SDLNet_SocketSet socketSet = NULL;
TCPsocket client[MAX_SNAKES] = {NULL, NULL, NULL, NULL};

void menuServer();
void gameServer();

static inline void closeClient(int i) {
    if (client[i]) {
	SDLNet_TCP_DelSocket(socketSet, client[i]);
	SDLNet_TCP_Close(client[i]);
	client[i] = NULL;
	remoteSnakes[i] = ai_remote;
	snprintf(names[i], 20, "Snake %i", i + 1);
    }
}

int server(void* data) {
    for (int i = 0; i < MAX_SNAKES; i++)
	closeClient(i);

    if (socketSet)
	SDLNet_FreeSocketSet(socketSet);

    socketSet = SDLNet_AllocSocketSet(MAX_SNAKES);

    if (!socketSet) {
	fprintf(stderr, "Connection error: %s\n", SDLNet_GetError());
	networkError = true;
	return 2;
    }
    
    while (!networkError) {
	switch (p_mode) {
	  case pm_menu: menuServer(); break;
	  case pm_game: gameServer(); break;
	  case pm_quit: networkError = true; break;
	}
    }

    for (int i = 0; i < MAX_SNAKES; i++)
	closeClient(i);
    SDLNet_FreeSocketSet(socketSet);
    socketSet = NULL;
    
    return 0;
}

void serverSendSnakeConfig(int snake) {
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i] && i != snake)
	    sendPacket_SnakeConfig(client[i], snake);
}

void serverSendBoardConfig() {
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i])
	    sendPacket_BoardConfig(client[i]);
}

void serverSendStart() {
    for (int i = 0; i < MAX_SNAKES; i++) {
	recvTicks[i] = false;
	if (client[i])
	    sendPacket_Start(client[i]);
    }
}

void serverSendTick() {
    //printf("server sent tick\n");
    for (int i = 0; i < MAX_SNAKES; i++) {
	//recvTicks[i] = false;
	if (client[i])
	    sendPacket_Tick(client[i]);
    }
}

void serverSendStop() {
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i])
	    sendPacket_Stop(client[i]);
}

void serverSendChangeDir(int snake, direction dir) {
    //printf("server sent change dir for %i\n", snake);
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i] && i != snake)
	    sendPacket_ChangeDir(client[i], snake, dir);
}

void serverSendInc(int snake) {
    //printf("server sent inc for %i\n", snake);
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i])
	    sendPacket_Inc(client[i], snake);
}

void serverSendFood() {
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i])
	    sendPacket_Food(client[i]);
}

void serverSendSpeed() {
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i])
	    sendPacket_Speed(client[i]);
}

void serverSendReset() {
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i])
	    sendPacket_Reset(client[i]);
}

void serverSendFineTune(int snake, unsigned char amount) {
    sendPacket_FineTune(client[snake], amount);
}

void serverSendPause(bool pause) {
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i])
	    sendPacket_Pause(client[i], pause);
}

void serverSendToggleFineTune(bool fineTune) {
    for (int i = 0; i < MAX_SNAKES; i++)
	if (client[i])
	    sendPacket_FineTune(client[i], fineTune);    
}

void menuServer() {
    IPaddress ip;
    ip.host = INADDR_ANY;
    ip.port = myIP.port;

    TCPsocket listener = SDLNet_TCP_Open(&ip);
    if (!listener) {
	fprintf(stderr, "Connection error: %s\n", SDLNet_GetError());
	networkError = true;
	return;
    }
    
    SDLNet_TCP_AddSocket(socketSet, listener);
    
    while (p_mode == pm_menu && !networkError) {
	
	int active = SDLNet_CheckSockets(socketSet, BASE_INTERVAL * 10);
	if (active < 0) {
	    networkError = true;
	    break;
	}
	else if (active > 0) {
	    DEBUG_PRINT("server activity: menu mode");
	    
	    // new connection
	    if (SDLNet_SocketReady(listener)) {
		DEBUG_PRINT(": new connection");
		int i = 0;
		for (i = 0; i < MAX_SNAKES; i++)
		    if (prefs.ai[i] == ai_remote && remoteSnakes[i] == ai_remote) break;
		if (i < MAX_SNAKES) {
		    client[i] = SDLNet_TCP_Accept(listener);
		    SDLNet_TCP_AddSocket(socketSet, client[i]);
		    remoteSnakes[i] = ai_human;
		    if (!sendPacket_Assign(client[i], i)) {
			closeClient(i);
		    } else {
			for (int j = 0; j < MAX_SNAKES; j++)
			    if (j != i)
				sendPacket_SnakeConfig(client[i], j);
			//rebuildBoard();
		    }
		    sendPacket_Speed(client[i]);
		    sendPacket_ToggleFineTune(client[i], prefs.fineTune);
		} else
		    SDLNet_TCP_Close(SDLNet_TCP_Accept(listener));
	    }

	    // remote update
	    else {
		for (int i = 0; i < MAX_SNAKES; i++) {
		    if (client[i] && SDLNet_SocketReady(client[i])) {
			DEBUG_PRINT2(": client %i", i);
			unsigned char packetType;
			//SDL_mutexP(netLock);
                        SDL_SemWait(netLock);
			int received = SDLNet_TCP_Recv(client[i], &packetType, 1);
			//SDL_mutexV(netLock);
                        SDL_SemPost(netLock);
			if (received < 0) {
			    DEBUG_PRINT2(": error - %s\n", SDLNet_GetError());
                            perror("Network error");
			    closeClient(i);
			    for (int j = 0; j < MAX_SNAKES; j++)
				if (client[j])
				    sendPacket_SnakeConfig(client[j], i);
			    rebuildBoard = true;
			} else if (received == 0) {
			    DEBUG_PRINT(": closed connection");
			    closeClient(i);
			    for (int j = 0; j < MAX_SNAKES; j++)
				if (client[j])
				    sendPacket_SnakeConfig(client[j], i);
			    rebuildBoard = true;
			} else {
			    
			    switch (packetType) {
			      case SNAKE_CONFIG_PACKET:
				  DEBUG_PRINT(": snake config");
				  unsigned char snake;
				  recvPacket_SnakeConfig(client[i], &snake);
				  if (i != snake) {
				      closeClient(i);
				  }
				  for (int j = 0; j < MAX_SNAKES; j++)
				      if (client[j] && j != i)
					  sendPacket_SnakeConfig(client[j], i);
				  rebuildBoard = true;
 				  break;

			      case START_PACKET:
				  DEBUG_PRINT(": start");
				  recvPacket_Start(client[i]);
				  recvTicks[i] = true;
				  break;
				  
			      case TICK_PACKET:
				  DEBUG_PRINT(": tick");
				  recvPacket_Tick(client[i]);
				  recvTicks[i] = true;
				  break;

			      case STOP_PACKET:
				  DEBUG_PRINT(": stop");
				  recvPacket_Stop(client[i]);
				  recvTicks[i] = false;
				  for (int j = 0; j < MAX_SNAKES; j++)
				      if (client[j] && j != i)
					  sendPacket_Stop(client[j]);
				  p_mode = pm_menu;
				  break;
				  
			      default:
				  DEBUG_PRINT(": error, unknown packet");
				  closeClient(i);
				  for (int j = 0; j < MAX_SNAKES; j++)
				      if (client[j])
					  sendPacket_SnakeConfig(client[j], i);
				  rebuildBoard = true;
			    }
			}
		    }
		}
	    }
	    DEBUG_PRINT("\n");
	}
	
	for (int i = 0; i < MAX_SNAKES; i++) {
	    if (remoteSnakes[i] == ai_remote && client[i]) {
		closeClient(i);
		for (int j = 0; j < MAX_SNAKES; j++)
		    if (client[j])
			sendPacket_SnakeConfig(client[j], i);
		rebuildBoard = true;
	    }
	}
    }
    
    SDLNet_TCP_DelSocket(socketSet, listener);
    SDLNet_TCP_Close(listener);
}

void gameServer() {
    while (p_mode == pm_game && !networkError) {
	int active = SDLNet_CheckSockets(socketSet, BASE_INTERVAL * 10);
	if (active < 0) {
	    networkError = true;
	    break;
	}
	else if (active > 0) {
	    DEBUG_PRINT("server activity: game mode");
	    
	    // remote update
	    for (int i = 0; i < MAX_SNAKES; i++) {
		if (client[i] && SDLNet_SocketReady(client[i])) {
		    DEBUG_PRINT2(": client %i", i);
		    unsigned char packetType;
		    //SDL_mutexP(netLock);
                    SDL_SemWait(netLock);
		    int received = SDLNet_TCP_Recv(client[i], &packetType, 1);
		    //SDL_mutexV(netLock);
                    SDL_SemPost(netLock);
		    if (received < 0) {
			DEBUG_PRINT2(": error - %s\n", SDLNet_GetError());
                        perror("Network error");
			closeClient(i);
			for (int j = 0; j < MAX_SNAKES; j++)
			    if (client[j])
				sendPacket_SnakeConfig(client[j], i);
			serverSendStop();
			p_mode = pm_menu;
		    } else if (received == 0) {
			DEBUG_PRINT(": closed connection");
			closeClient(i);
			for (int j = 0; j < MAX_SNAKES; j++)
			    if (client[j])
				sendPacket_SnakeConfig(client[j], i);
			serverSendStop();
			p_mode = pm_menu;
		    } else {
 
			switch (packetType) {
				  
			  case TICK_PACKET:
			      DEBUG_PRINT(": tick");
			      recvPacket_Tick(client[i]);
			      if (!reseting || recvReset[i])
				  recvTicks[i] = true;
			      break;
			      
			  case CHANGE_DIR_PACKET:
			      DEBUG_PRINT(": change direction");
			      unsigned char snake;
			      direction dir;
			      recvPacket_ChangeDir(client[i], &snake, &dir);
			      //while (recvTicks[i]);
			      //SDL_mutexP(stepLock);
			      if (!reseting || recvReset[i]) {
				  //SDL_SemWait(stepLock);
				  lockStep("got change dir");
				  snakes[snake]->changeDirection(dir);
				  //SDL_mutexV(stepLock);
				  //SDL_SemPost(stepLock);
				  unlockStep("end change dir");
				  serverSendChangeDir(snake, dir);
			      }
			      break;

			  case STOP_PACKET:
			      DEBUG_PRINT(": stop");
			      recvPacket_Stop(client[i]);
			      for (int j = 0; j < MAX_SNAKES; j++)
				  if (client[j] && j != i)
				      sendPacket_Stop(client[j]);
			      p_mode = pm_menu;
			      break;
			      
			  case RESET_PACKET:
			      DEBUG_PRINT(": reset");
			      recvPacket_Reset(client[i]);
			      recvReset[i] = true;
			      break;

			  case PAUSE_PACKET:
			      DEBUG_PRINT(": pause");
			      unsigned char rpause;
			      recvPacket_Pause(client[i], &rpause);
			      paused = rpause;
			      for (int j = 0; j < MAX_SNAKES; j++)
				  if (client[j] && j != i)
				      sendPacket_Pause(client[j], rpause);
			      break;
			      
			  default:
			      DEBUG_PRINT(": error, unknown packet");
			      closeClient(i);
			      for (int j = 0; j < MAX_SNAKES; j++)
				  if (client[j])
				      sendPacket_SnakeConfig(client[j], i);
			      serverSendStop();
			      p_mode = pm_menu;
			}
		    }
		}
	    }
	    DEBUG_PRINT("\n");
	}
	
    }
}
