/*  
 *  snake3d, a snake game in a 3d world.
 *  Copyright (C) 2005-2006 Yair Yarom
 *  
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or (at your option) any later
 *  version.
 *  
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License along with
 *  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 *  Place, Suite 330, Boston, MA 02111-1307 USA
 *  
 *  http://sourceforge.net/projects/worms3d
 *  
 */
 
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "snake.hpp"
#include "board.hpp"
#include "consts.hpp"
#include "global.hpp"
#include "functions.hpp"
#include "preferences.hpp"

Snake::Snake(int x, int y, int z, direction dir, direction up, int color, int location): _startDir(dir), _startUp(up), _id(location) {
    _joints = new int*[MAX_SNAKE_LENGTH];
    for (int i = 0; i < MAX_SNAKE_LENGTH; i++)
        _joints[i] = new int[6];
    
    _start[X] = x;
    _start[Y] = y;
    _start[Z] = z;
    _length = STARTING_SNAKE_LENGTH;
    _autoPilot = ai_off;
    _aiTarget = ait_food;
    _aitSnake = 0;
    _color = color;
    reset();
    
    //_currentLength = 1;
    //_joints[0][X] = x;
    //_joints[0][Y] = y;
    //_joints[0][Z] = z;
    //_joints[0][HAS_FOOD] = false;
    //_joints[0][ENTER] = dir;
    //_joints[0][EXIT] = dir;
    //_dir = dir;
    //_oldDir = dir;
    //_up = STARTING_UP[_dir];
    //_oldUp = _up;
    //_dead = false;
}

void Snake::reset() {
    _currentLength = 1;
    _joints[0][X] = _start[X];
    _joints[0][Y] = _start[Y];
    _joints[0][Z] = _start[Z];
    _joints[0][HAS_FOOD] = false;
    _joints[0][ENTER] = _startDir;
    _joints[0][EXIT] = _startDir;
    _joints[1][X] = _joints[0][X];
    _joints[1][Y] = _joints[0][Y];
    _joints[1][Z] = _joints[0][Z];
    switch (_startDir) {
      case left:     _joints[1][X]++; break;
      case up:       _joints[1][Y]--; break;
      case right:    _joints[1][X]--; break;
      case down:     _joints[1][Y]++; break;
      case forward:  _joints[1][Z]--; break;
      case backward: _joints[1][Z]++; break;
    }
    _dir = _startDir;
    _up = _startUp;
    _dead = false;
    //_longDead = false;
    _oldDir = _dir;
    _oldUp = _up;
}

void Snake::kill() {
    _dead = true;
    _joints[_currentLength - 1][HAS_FOOD] = false;
}

bool Snake::isDead() {
    return _dead;
}

occupancy_t Snake::collision(int* location, bool me) {
    //for (int i = (me || _dead); i < _currentLength; i++)
    for (int i = me; i < _currentLength; i++)
        if (location[X] == _joints[i][X] &&
            location[Y] == _joints[i][Y] &&
            location[Z] == _joints[i][Z])
            return o_snake;
    
    int next[3];
    step(_joints[0], _dir, next);
    if (!me && !_dead && next[X] == location[X] && next[Y] == location[Y] && next[Z] == location[Z])
	return o_intercept;
    return o_empty;
}

void Snake::move() {
    if (_dead)
        return;

    //changeDirection(_dir);

    // move
    if (_joints[_currentLength - 1][HAS_FOOD]) {
        if (_length < MAX_SNAKE_LENGTH) _length++;
        _joints[_currentLength - 1][HAS_FOOD] = false;
    }
    if (_currentLength < _length) {
        _currentLength++;
    }
    int* head = _joints[_currentLength - 1];
    for (int i = _currentLength - 1; i > 0; i--)
        _joints[i] = _joints[i - 1];
    _joints[0] = head;
    head[HAS_FOOD] = false;
    step(_joints[1], _dir, head);
    _joints[1][ENTER] = _oldDir;
    _joints[1][EXIT] = _dir;
    _oldDir = _dir;
    _oldUp = _up;
}

void Snake::autoPilot(ai_t apMode, aiTarget_t aiTarget, Snake* aitSnake) {
    if (apMode == ai_switch)
        _autoPilot = (_autoPilot == ai_human ? ai_safe : ai_human);
    else
        _autoPilot = apMode;
    _aiTarget = aiTarget;
    _aitSnake = aitSnake;
}

void Snake::inc() {
    _joints[0][HAS_FOOD] = true;
    if (networkState == net_server)
	sendInc(_id);
}

void Snake::changeDirection(direction dir) {
    if (_dead)
	return;

    int prevDir = _dir;

    int target[3];
    bool hasTarget = false;
    switch (_autoPilot) {
      case ai_safe:
	  if (dir != _dir)
	      _autoPilot = ai_human;
	  else {
	      step(_joints[0], _dir, target);
	      hasTarget = true;
	      break;
	  }
	  
      case ai_human:
	  if (_oldDir == OPPOSITE[dir])
	      return;
	  if (_dir != _oldDir) {
	      _up = _oldUp;
	      _dir = _oldDir;
	  }
	  _up = ROTATING_UP[_dir][dir][_up];
	  _dir = dir;
	  break;
	  
      case ai_basic:
	  hasTarget = true;
	  switch (_aiTarget) {
	    case ait_food:
		target[X] = mainBoard->getFood()[X];
		target[Y] = mainBoard->getFood()[Y];
		target[Z] = mainBoard->getFood()[Z];
		break;

	    case ait_head:
		target[X] = _aitSnake->_joints[0][X];
		target[Y] = _aitSnake->_joints[0][Y];
		target[Z] = _aitSnake->_joints[0][Z];
		break;
		
	    case ait_tail:
		target[X] = _aitSnake->_joints[_aitSnake->_currentLength - 1][X];
		target[Y] = _aitSnake->_joints[_aitSnake->_currentLength - 1][Y];
		target[Z] = _aitSnake->_joints[_aitSnake->_currentLength - 1][Z];
		break;
	  }
	  break;

      case ai_remote:
	  if (_dir != _oldDir) {
	      _up = _oldUp;
	      _dir = _oldDir;
	  }
	  _up = ROTATING_UP[_dir][dir][_up];
	  _dir = dir;
	  //printf("changed direction in remote ai\n");
	  break;
	  
      case ai_switch:
      case ai_off:
	  quit(1);
	  ;
    }

    if (hasTarget) {
        int posibility[3];
        double currentDistance = distance(_joints[0], target);
        step(_joints[0], _dir, posibility);
        double nextDistance = distance(posibility, target);
        occupancy_t occupancy = mainBoard->checkLocation(posibility, this);
        if (currentDistance < nextDistance || deadEnd(occupancy)) {
	    direction newDir = _dir;
	    double best = mainBoard->getMaxSize() * 5;
	    occupancy_t bestOccupancy = occupancy;
	    const direction dirs[6] = {left, right, up, down, forward, backward};
	    for (int i = 0; i < 6; i++) {
		step(_joints[0], dirs[i], posibility);
		nextDistance = distance(posibility, target);
		occupancy = mainBoard->checkLocation(posibility, this);
		//if (nextDistance < best && (!deadEnd(occupancy) || (occupancy == o_intercept && (((double)random() / RAND_MAX) < 0.25)))) {
		if (nextDistance < best && !deadEnd(occupancy) && dirs[i] != OPPOSITE[_oldDir]) {
		    newDir = dirs[i];
		    best = nextDistance;
		    bestOccupancy = occupancy;
		} else if (occupancy == o_intercept && deadEnd(bestOccupancy)) {
		    newDir = dirs[i];
		    bestOccupancy = occupancy;
		}
	    }
	    _up = ROTATING_UP[_dir][newDir][_up];
            _dir = newDir;
        }
    }

    if (networkState != net_none && _autoPilot != ai_remote && prevDir != _dir)
	sendChangeDir(_id, _dir);
}

void Snake::changeSelfDirection(direction dir) {
    changeDirection(SELF_ROTATING[_oldDir][dir][_oldUp]);
}

void Snake::changeYaw(direction dir) {
    if (_autoPilot == ai_safe) 
	_autoPilot = ai_human;
    _up = YAW_UP[_oldDir][_oldUp][dir != left];
    _dir = _oldDir;
    _oldUp = _up;
}

int Snake::getColor(){
    return _color;
}

int Snake::getLength(){
    return _length;
}

inline void drawBend(direction enter, direction exit, float percent) {
    /*
    if (enter == -(X + 1) || exit == -(X + 1)) {x1 = Y; y1 = X; z1 = Z;}//glVertex3f(-0.5, 0, 0);
    if (enter == -(Y + 1) || exit == -(Y + 1)) {x1 = X; y1 = Y; z1 = Z;}//glVertex3f(0, -0.5, 0);
    if (enter == -(Z + 1) || exit == -(Z + 1)) {x1 = X; y1 = Z; z1 = Y;}//glVertex3f(0, 0, -0.5);
    if (enter == (X + 1) || exit == (X + 1)) {x2 = Y; y2 = X; z2 = Z;}//glVertex3f(0.5, 0, 0);
    if (enter == (Y + 1) || exit == (Y + 1)) {x2 = X; y2 = Y; z2 = Z;}//glVertex3f(0, 0.5, 0);
    if (enter == (Z + 1) || exit == (Z + 1)) {x2 = X; y2 = Z; z2 = Y;}//glVertex3f(0, 0, 0.5);
    */

    switch (enter) {
        case up:
            glScalef(1, -1, 1);
        case down:
            switch (exit) {
                case right:    glRotatef(180, 0, 1, 0); break;
                case forward:  glRotatef(-90, 0, 1, 0); break;
                case backward: glRotatef( 90, 0, 1, 0); break;
                default:;
            }
            break;
        case right:
            glScalef(-1, 1, 1);
        case left:
            switch (exit) {
                case down:     glRotatef(180, 1, 0, 0); break;
                case forward:  glRotatef(-90, 1, 0, 0); break;
                case backward: glRotatef( 90, 1, 0, 0); break;
                default:;
            }
            glRotatef(90, 0, 0, -1);
            break;
        case backward:
            glScalef(1, 1, -1);
        case forward:
            switch (exit) {
                case up:       glRotatef(-90, 0, 0, 1); break;
                case down:     glRotatef( 90, 0, 0, 1); break;
                case right:    glRotatef(180, 0, 0, 1); break;
                default:;
            }
            glRotatef(90, 1, 0, 0);
            break;
        default:;
    }

    
    double edge[9][4];
    double edgeNormal[9][4];

    for (int i = 0; i < 9; i++) {
        edge[i][X] = (0.5 + JOINT_EDGES[i][X]) * cos(percent * M_PI / 2) - 0.5;
	edge[i][Y] = (0.5 + JOINT_EDGES[i][X]) * sin(percent * M_PI / 2) - 0.5;
        edge[i][Z] = JOINT_EDGES[i][Z];

        edgeNormal[i][X] = (0.5 + JOINT_NORMALS[i][X]) * cos(percent * M_PI / 2) - 0.5;
	edgeNormal[i][Y] = (0.5 + JOINT_NORMALS[i][X]) * sin(percent * M_PI / 2) - 0.5;
        edgeNormal[i][Z] = JOINT_NORMALS[i][Z];
    }


    glBegin(GL_QUADS);
    for (int i = 0; i < 8; i++) {
	
	//glBegin(GL_LINE_LOOP);
	glNormal3f(JOINT_NORMALS[i][X], JOINT_NORMALS[i][Y], JOINT_NORMALS[i][Z]);
	glVertex3f(JOINT_EDGES[i][X], JOINT_EDGES[i][Y], JOINT_EDGES[i][Z]);
	glNormal3f(JOINT_NORMALS[i + 1][X], JOINT_NORMALS[i + 1][Y], JOINT_NORMALS[i + 1][Z]);
	glVertex3f(JOINT_EDGES[i + 1][X], JOINT_EDGES[i + 1][Y], JOINT_EDGES[i + 1][Z]);
	if (percent < 1.0 / 3.0) {
	    
	    glNormal3f(edgeNormal[i+1][X], edgeNormal[i+1][Y], edgeNormal[i+1][Z]);
	    glVertex3f(edge[i+1][X], edge[i+1][Y], edge[i+1][Z]);
	    glNormal3f(edgeNormal[i][X], edgeNormal[i][Y], edgeNormal[i][Z]);
	    glVertex3f(edge[i][X], edge[i][Y], edge[i][Z]);
	    //glEnd();
	    
	} else {
	    glNormal3f(BEND_NORMALS[i + 1][X], BEND_NORMALS[i + 1][Y], BEND_NORMALS[i + 1][Z]);
	    glVertex3f(JOINT_BENDS[i + 1][X], JOINT_BENDS[i + 1][Y], JOINT_BENDS[i + 1][Z]);
	    glNormal3f(BEND_NORMALS[i][X], BEND_NORMALS[i][Y], BEND_NORMALS[i][Z]);
	    glVertex3f(JOINT_BENDS[i][X], JOINT_BENDS[i][Y], JOINT_BENDS[i][Z]);
	    //glEnd();

	    //glBegin(GL_LINE_LOOP);
	    glNormal3f(BEND_NORMALS[i][X], BEND_NORMALS[i][Y], BEND_NORMALS[i][Z]);
	    glVertex3f(JOINT_BENDS[i][X], JOINT_BENDS[i][Y], JOINT_BENDS[i][Z]);
	    glNormal3f(BEND_NORMALS[i + 1][X], BEND_NORMALS[i + 1][Y], BEND_NORMALS[i + 1][Z]);
	    glVertex3f(JOINT_BENDS[i + 1][X], JOINT_BENDS[i + 1][Y], JOINT_BENDS[i + 1][Z]);
	    if (percent < 2.0 / 3.0) {
		
		glNormal3f(edgeNormal[i+1][X], edgeNormal[i+1][Y], edgeNormal[i+1][Z]);
		glVertex3f(edge[i+1][X], edge[i+1][Y], edge[i+1][Z]);
		glNormal3f(edgeNormal[i][X], edgeNormal[i][Y], edgeNormal[i][Z]);
		glVertex3f(edge[i][X], edge[i][Y], edge[i][Z]);		
		//glEnd();
		
	    } else {
		glNormal3f(BEND_NORMALS[i + 1][Y], BEND_NORMALS[i + 1][X], BEND_NORMALS[i + 1][Z]);
		glVertex3f(JOINT_BENDS[i + 1][Y], JOINT_BENDS[i + 1][X], JOINT_BENDS[i + 1][Z]);
		glNormal3f(BEND_NORMALS[i][Y], BEND_NORMALS[i][X], BEND_NORMALS[i][Z]);
		glVertex3f(JOINT_BENDS[i][Y], JOINT_BENDS[i][X], JOINT_BENDS[i][Z]);
		//glEnd();

		//glBegin(GL_LINE_LOOP);
		glNormal3f(BEND_NORMALS[i + 1][Y], BEND_NORMALS[i + 1][X], BEND_NORMALS[i + 1][Z]);
		glVertex3f(JOINT_BENDS[i + 1][Y], JOINT_BENDS[i + 1][X], JOINT_BENDS[i + 1][Z]);
		glNormal3f(BEND_NORMALS[i][Y], BEND_NORMALS[i][X], BEND_NORMALS[i][Z]);
		glVertex3f(JOINT_BENDS[i][Y], JOINT_BENDS[i][X], JOINT_BENDS[i][Z]);
		if (percent < 1.0) {

		    glNormal3f(edgeNormal[i][X], edgeNormal[i][Y], edgeNormal[i][Z]);
		    glVertex3f(edge[i][X], edge[i][Y], edge[i][Z]);		
		    glNormal3f(edgeNormal[i+1][X], edgeNormal[i+1][Y], edgeNormal[i+1][Z]);
		    glVertex3f(edge[i+1][X], edge[i+1][Y], edge[i+1][Z]);
		    //glEnd();
		} else {
		    		    
		    glNormal3f(JOINT_NORMALS[i][Y], JOINT_NORMALS[i][X], JOINT_NORMALS[i][Z]);
		    glVertex3f(JOINT_EDGES[i][Y], JOINT_EDGES[i][X], JOINT_EDGES[i][Z]);		    
		    glNormal3f(JOINT_NORMALS[i + 1][Y], JOINT_NORMALS[i + 1][X], JOINT_NORMALS[i + 1][Z]);
		    glVertex3f(JOINT_EDGES[i + 1][Y], JOINT_EDGES[i + 1][X], JOINT_EDGES[i + 1][Z]);		    
		    //glEnd();
		}
	    }
	}
    }
    glEnd();
}

inline void headRotation(direction dir, direction Up) {
    switch (dir) {
      case left:
            switch (Up) {
                case down:     glRotatef(180, 1, 0, 0); break;
                case forward:  glRotatef(-90, 1, 0, 0); break;
                case backward: glRotatef( 90, 1, 0, 0); break;
                default:                                break;
            }
            break;

      case down:
            switch (Up) {
                case right:    glRotatef(180, 0, 1, 0); break;
                case forward:  glRotatef(-90, 0, 1, 0); break;
                case backward: glRotatef( 90, 0, 1, 0); break;
                default:                                break;
            }
            glRotatef( 90, 0, 0, 1);
            break;

      case right:
            switch (Up) {
                case down:     glRotatef(180, 1, 0, 0); break;
                case forward:  glRotatef(-90, 1, 0, 0); break;
                case backward: glRotatef( 90, 1, 0, 0); break;
                default:                                break;
            }
            glRotatef(180, 0, 1, 0);
            break;

      case up:
            switch (Up) {
                case left:     glRotatef(180, 0, 1, 0); break;
                case forward:  glRotatef( 90, 0, 1, 0); break;
                case backward: glRotatef(-90, 0, 1, 0); break;
                default:                                break;
            }
            glRotatef(-90, 0, 0, 1);
            break;

      case forward:
	  switch (Up) {
	    case down:     glRotatef(180, 0, 0, 1); break;
	    case left:     glRotatef( 90, 0, 0, 1); break;
	    case right:    glRotatef(-90, 0, 0, 1); break;
	    default:                                break;
	  }
	  glRotatef(-90, 0, 1, 0);
	  break;

      case backward:
            switch (Up) {
                case down:     glRotatef(180, 0, 0, 1); break;
                case left:     glRotatef( 90, 0, 0, 1); break;
                case right:    glRotatef(-90, 0, 0, 1); break;
                default:                                break;
            }
            glRotatef( 90, 0, 1, 0);
            break;
    }    
};

inline void headBend(direction oldDir, direction newDir, float percent) {
    percent *= 90;
    glTranslatef(HEAD_BEND_BASE[oldDir][0],
		 HEAD_BEND_BASE[oldDir][1],
		 HEAD_BEND_BASE[oldDir][2]);
    
    glTranslatef(       HEAD_BEND[oldDir][newDir][0][0],
			HEAD_BEND[oldDir][newDir][0][1],
			HEAD_BEND[oldDir][newDir][0][2]);
    
    glRotatef(percent,  HEAD_BEND[oldDir][newDir][1][0], 
  	                HEAD_BEND[oldDir][newDir][1][1],
	                HEAD_BEND[oldDir][newDir][1][2]);
    
    glTranslatef(      -HEAD_BEND[oldDir][newDir][0][0], 
		       -HEAD_BEND[oldDir][newDir][0][1], 
		       -HEAD_BEND[oldDir][newDir][0][2]);
    
    /*switch (oldDir) {
        case left:
            glTranslatef(0.5, 0, 0);
            switch (newDir) {
                case backward:
                    glTranslatef(0, 0,  0.5);
                    glRotatef(percent, 0, 1, 0);
                    glTranslatef(0, 0, -0.5);
                    break;
                case forward:
                    glTranslatef(0, 0, -0.5);
                    glRotatef(percent, 0, -1, 0);
                    glTranslatef(0, 0,  0.5);
                    break;
                case up:
                    glTranslatef(0,  0.5, 0);
                    glRotatef(percent, 0, 0, -1);
                    glTranslatef(0, -0.5, 0);
                    break;
                case down:
                    glTranslatef(0, -0.5, 0);
                    glRotatef(percent, 0, 0, 1);
                    glTranslatef(0,  0.5, 0);
                    break;
                default:;
            };
            break;
        case right:
            glTranslatef(-0.5, 0, 0);
            switch (newDir) {
                case backward:
                    glTranslatef(0, 0,  0.5);
                    glRotatef(percent, 0, -1, 0);
                    glTranslatef(0, 0, -0.5);
                    break;
                case forward:
                    glTranslatef(0, 0, -0.5);
                    glRotatef(percent, 0, 1, 0);
                    glTranslatef(0, 0, 0.5);
                    break;
                case up:
                    glTranslatef(0,  0.5, 0);
                    glRotatef(percent, 0, 0, 1);
                    glTranslatef(0, -0.5, 0);
                    break;
                case down:
                    glTranslatef(0, -0.5, 0);
                    glRotatef(percent, 0, 0, -1);
                    glTranslatef(0, 0.5, 0);
                    break;
                default:;
            };
            break;
        case forward:
            glTranslatef(0, 0, 0.5);
            switch (newDir) {
                case left:
                    glTranslatef(-0.5, 0, 0);
                    glRotatef(percent, 0, 1, 0);
                    glTranslatef(0.5, 0, 0);
                    break;
                case right:
                    glTranslatef( 0.5, 0, 0);
                    glRotatef(percent, 0, -1, 0);
                    glTranslatef(-0.5, 0, 0);
                    break;
                case up:
                    glTranslatef(0,  0.5, 0);
                    glRotatef(percent, 1, 0, 0);
                    glTranslatef(0, -0.5, 0);
                    break;
                case down:
                    glTranslatef(0, -0.5, 0);
                    glRotatef(percent, -1, 0, 0);
                    glTranslatef(0,  0.5, 0);
                    break;
                default:;
            };
            break;
        case backward:
            glTranslatef(0, 0, -0.5);
            switch (newDir) {
                case left:
                    glTranslatef(-0.5, 0, 0);
                    glRotatef(percent, 0, -1, 0);
                    glTranslatef(0.5, 0, 0);
                    break;
                case right:
                    glTranslatef( 0.5, 0, 0);
                    glRotatef(percent, 0, 1, 0);
                    glTranslatef(-0.5, 0, 0);
                    break;
                case up:
                    glTranslatef(0,  0.5, 0);
                    glRotatef(percent, -1, 0, 0);
                    glTranslatef(0, -0.5, 0);
                    break;
                case down:
                    glTranslatef(0, -0.5, 0);
                    glRotatef(percent, 1, 0, 0);
                    glTranslatef(0,  0.5, 0);
                    break;
                default:;
            };
            break;
        case up:
            glTranslatef(0, -0.5, 0);
            switch (newDir) {
                case left:
                    glTranslatef(-0.5, 0, 0);
                    glRotatef(percent, 0, 0, 1);
                    glTranslatef( 0.5, 0, 0);
                    break;
                case right:
                    glTranslatef( 0.5, 0, 0);
                    glRotatef(percent, 0, 0, -1);
                    glTranslatef(-0.5, 0, 0);
                    break;
                case forward:
                    glTranslatef(0, 0, -0.5);
                    glRotatef(percent, -1, 0, 0);
                    glTranslatef(0, 0,  0.5);
                    break;
                case backward:
                    glTranslatef(0, 0,  0.5);
                    glRotatef(percent, 1, 0, 0);
                    glTranslatef(0, 0, -0.5);
                    break;
                default:;
            };
            break;
        case down:
            glTranslatef(0, 0.5, 0);
            switch (newDir) {
                case left:
                    glTranslatef(-0.5, 0, 0);
                    glRotatef(percent, 0, 0, -1);
                    glTranslatef( 0.5, 0, 0);
                    break;
                case right:
                    glTranslatef( 0.5, 0, 0);
                    glRotatef(percent, 0, 0, 1);
                    glTranslatef(-0.5, 0, 0);
                    break;
                case forward:
                    glTranslatef(0, 0, -0.5);
                    glRotatef(percent, 1, 0, 0);
                    glTranslatef(0, 0,  0.5);
                    break;
                case backward:
                    glTranslatef(0, 0,  0.5);
                    glRotatef(percent, -1, 0, 0);
                    glTranslatef(0, 0, -0.5);
                    break;
                default:;
            };
            break;
	    }*/
}


void Snake::drawJoint(int joint) {
    float color[4];
    glGetFloatv(GL_CURRENT_COLOR, color);
    glMaterialfv(GL_FRONT, GL_AMBIENT, SNAKE_AMBIENT);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
    glMaterialfv(GL_FRONT, GL_SPECULAR, SNAKE_SPECULAR);
    glMaterialf(GL_FRONT, GL_SHININESS, SNAKE_SHININESS);
    glMaterialfv(GL_FRONT, GL_EMISSION, SNAKE_EMISSION);
    
    glPushMatrix();
    //float percent = (float)(timerLong % SPEEDS[prefs.stepInterval]) / (SPEEDS[prefs.stepInterval] + 1);
    float percent = stepPercent;
    if (_dead) percent = 0.5;
    /*
    if (_dead)
	if (_longDead || percent > 0.5) {
	    percent = 0.5;
	    _longDead = true;
	}
    */
    
    glTranslated(_joints[joint][X] + 0.5, _joints[joint][Y] + 0.5, _joints[joint][Z] + 0.5);
    if (joint == 0) {
        glPushMatrix();
        //headRotation(_oldDir, _oldUp);
        
        // head
        if (_dir != _oldDir && OPPOSITE[_dir] != _oldDir) {
            glPushMatrix();
            headBend(_oldDir, _dir, percent);
            headRotation(_oldDir, _oldUp);
        } else {
            headRotation(_oldDir, _oldUp);
            glPushMatrix();
            glTranslatef(0.5 - percent , 0, 0);
        }
        glutSolidSphere(0.4, 8, 8);
	
        //glColor3fv(COLORS[BLACK]);
        //glutWireSphere(0.4, 8, 8);

	if (_dead) {
	    glMaterialfv(GL_FRONT, GL_DIFFUSE, COLORS[_color]);
	} else {
	    glMaterialfv(GL_FRONT, GL_AMBIENT, EYES_AMBIENT);
	    glMaterialfv(GL_FRONT, GL_DIFFUSE, EYES_DIFFUSE);
	    glMaterialfv(GL_FRONT, GL_SPECULAR, EYES_SPECULAR);
	    glMaterialf(GL_FRONT, GL_SHININESS, EYES_SHININESS);
	    glMaterialfv(GL_FRONT, GL_EMISSION, EYES_EMISSION);
        }
	
        // eyes
        glPushMatrix();
        glColor3fv((_dead ? COLORS[_color] : COLORS[YELLOW]));
        glTranslated(-0.24, 0.24, 0.19);
        glutSolidSphere(0.11, 8, 8);
        glTranslated(0, 0, -0.38);
        glutSolidSphere(0.11, 8, 8);
        glPopMatrix();
        
        glPopMatrix();

        glMaterialfv(GL_FRONT, GL_AMBIENT, SNAKE_AMBIENT);
        //glMaterialfv(GL_FRONT, GL_DIFFUSE, COLORS[_color]);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
        glMaterialfv(GL_FRONT, GL_SPECULAR, SNAKE_SPECULAR);
        glMaterialf(GL_FRONT, GL_SHININESS, SNAKE_SHININESS);
        glMaterialfv(GL_FRONT, GL_EMISSION, SNAKE_EMISSION);
        
        // tail
        //glPushMatrix();
        if (_dir != _oldDir && OPPOSITE[_dir] != _oldDir) {
            drawBend(OPPOSITE[_oldDir], _dir, percent);
        }
        else {
            glRotatef(-90, 0, 1, 0);
            glTranslated(0, 0, -0.5);
            glColor3fv(COLORS[_color]);
            gluQuadricDrawStyle(cylinder, GLU_FILL);
            gluCylinder(cylinder, 0.3, 0.3, percent, 8 , 1);
            //glColor3fv(COLORS[BLACK]);
            //gluQuadricDrawStyle(cylinder, GLU_LINE);
            //gluCylinder(cylinder, 0.3, 0.3, percent, 8, 1);
        }
        //glPopMatrix();

	glPopMatrix();
	
	// if just a head, put a cone tail
	if (_currentLength == 1) {
	    glPushMatrix();
            headRotation(_oldDir, _oldUp);	    
	    /*
	      switch (_oldDir) {
	      left:    glRotatef( -90, 0, 1, 0); break;
	      right:   glRotatef(  90, 0, 1, 0); break;
	      down:    glRotatef(  90, 1, 0, 0); break;
	      up:      glRotatef( -90, 1, 0, 0); break;
	      forward: glRotatef( 180, 0, 1, 0); break;
	      default:;
	      }
	      glRotatef(90, 0, 1, 0);
	      glTranslated(0, 0, 0.5);
	    */
	    
	    
	    glRotatef(90, 0, 1, 0);
	    glTranslatef(0, 0, 0.5);
	    
	    glutSolidCone(0.3, 0.5, 8, 1);
	    glPopMatrix();
	}

        if (_joints[joint][HAS_FOOD]) {
            glColor3fv(COLORS[_color]);
            glutSolidSphere(percent * 0.39, 8, 8);
        }
	
    } else if  (joint == _currentLength - 1) {
	//} else if (false) {


        if (_joints[joint][HAS_FOOD]) {
            glutSolidSphere(0.39 * (1 - percent), 8, 8);
            if (_length < MAX_SNAKE_LENGTH) percent = 0;
        }
        if (_currentLength != _length)
            percent = 0;

	if (_joints[joint][ENTER] != _joints[joint][EXIT]) {

	    glPushMatrix();
	    drawBend((direction)_joints[joint][EXIT], OPPOSITE[_joints[joint][ENTER]], 1 - percent);
	    glPopMatrix();
	    
	    headBend((direction)_joints[joint][ENTER], (direction)_joints[joint][EXIT], percent);
	    if      (_joints[joint][ENTER] == right)   	glRotatef(-90, 0, 1, 0);
	    else if (_joints[joint][ENTER] == left)    	glRotatef( 90, 0, 1, 0);
	    else if (_joints[joint][ENTER] == up)      	glRotatef( 90, 1, 0, 0);
	    else if (_joints[joint][ENTER] == down)    	glRotatef(-90, 1, 0, 0);
	    else if (_joints[joint][ENTER] == backward) glRotatef(180, 0, 1, 0);
	    glRotatef(90, 0, 0, 1);
	    
	    glutSolidCone(0.3, 0.5, 8, 1);
	    
	} else {
	    if      (_joints[joint][ENTER] == right)   	glRotatef(-90, 0, 1, 0);
	    else if (_joints[joint][ENTER] == left)    	glRotatef( 90, 0, 1, 0);
	    else if (_joints[joint][ENTER] == up)      	glRotatef( 90, 1, 0, 0);
	    else if (_joints[joint][ENTER] == down)    	glRotatef(-90, 1, 0, 0);
	    else if (_joints[joint][ENTER] == backward) glRotatef(180, 0, 1, 0);

	    glPushMatrix();
	    gluQuadricDrawStyle(cylinder, GLU_FILL);
	    glTranslated(0, 0, -0.5);
	    gluCylinder(cylinder, 0.3, 0.3, 1 - percent, 8 ,1);
	    
	    //glColor3fv(COLORS[BLACK]);
	    //gluQuadricDrawStyle(cylinder, GLU_LINE);
	    //gluCylinder(cylinder, 0.3, 0.3, 1 - percent, 8, 1);
	    glPopMatrix();
	    
	    //glColor3fv(COLORS[_color]);
	    glTranslated(0 , 0 , 0.5 - percent);
	    glutSolidCone(0.3, 0.5, 8, 1);
	    //glColor3fv(COLORS[BLACK]);
	    //glutWireCone(0.3, 0.5, 8, 1);
	}
    } else {

	if (_joints[joint][ENTER] == _joints[joint][EXIT] && (_joints[joint][ENTER] == left || _joints[joint][ENTER] == right))
            glRotatef(90, 0, 1, 0);
	else if (_joints[joint][ENTER] == _joints[joint][EXIT] && (_joints[joint][ENTER] == up || _joints[joint][ENTER] == down))
            glRotatef(90, 1, 0, 0);

	if (_joints[joint][ENTER] != _joints[joint][EXIT]) {
            if (_joints[joint][HAS_FOOD])
                glutSolidSphere(0.3, 8, 8);
	    
	    drawBend(OPPOSITE[_joints[joint][ENTER]], (direction)_joints[joint][EXIT], 1.0);
        } else {
            if (_joints[joint][HAS_FOOD])
                glutSolidSphere(0.39, 8, 8);
            glTranslated(0, 0, -0.5);

            gluQuadricDrawStyle(cylinder, GLU_FILL);
            gluCylinder(cylinder, 0.3, 0.3, 1, 8, 1);

            //glColor3fv(COLORS[BLACK]);
            //gluQuadricDrawStyle(cylinder, GLU_LINE);
            //gluCylinder(cylinder, 0.3, 0.3, 1, 8, 1);
        }
    }
    
    glPopMatrix();
}

void Snake::draw(bool withHead) {
    //changeDirection(_dir);
    float color[3];
    if (_dead) {
        color[RED] = COLORS[RED][RED];
        color[GREEN] = COLORS[RED][GREEN];
        color[BLUE] = COLORS[RED][BLUE];
        //glColor3fv(color);
    } else {
        color[RED] = COLORS[_color][RED];
        color[GREEN] = COLORS[_color][GREEN];
        color[BLUE] = COLORS[_color][BLUE];
        //glColor3f(color[RED] - 0.2, color[GREEN] - 0.2, color[BLUE] - 0.2);
    }
    //for (int i = (_dead || withHead); i < _currentLength; i++) {
    for (int i = withHead; i < _currentLength; i++) {
        glColor3fv(color);
	glPushMatrix();
        drawJoint(i);
	glPopMatrix();
    }
}


void Snake::lookAt() {
    //float percent = (float)(timerLong % SPEEDS[prefs.stepInterval]) / (SPEEDS[prefs.stepInterval] + 1);
    float percent = stepPercent;
    if (_dead) percent = 0.5;
    //if (_dead && (_longDead || percent > 0.5)) percent = 0.5;
	    
    
    //percent -= 0.1;
    float cEye[3];
    float cUp[3];
    float cCenter[3];

    for (int i = X; i <= Z; i++) {
        float headDir;
	/*if (_currentLength == 1) {
	    headDir = 0;
	    if (i == X && (_dir == left    || _dir == right   )) headDir = (_dir == right   ) * 2 - 1;
	    if (i == Y && (_dir == up      || _dir == down    )) headDir = (_dir == up      ) * 2 - 1;
	    if (i == Z && (_dir == forward || _dir == backward)) headDir = (_dir == forward) * 2 - 1;
	    } else*/
	    headDir = _joints[0][i] - _joints[1][i];
        if (headDir && _dir == _oldDir) 
            cEye[i] = _joints[0][i] + headDir * percent;
        else
            cEye[i] = _joints[0][i] + 0.5;
        cUp[i] = UP_VECTOR[_oldUp][i];
        cCenter[i] = cEye[i] + headDir;
        if (headDir < 0)
            cEye[i]++;
    }

    if (_dir != _oldDir && OPPOSITE[_dir] != _oldDir) {
        switch(RELATIVE_ROTATION[_oldDir][_dir][_oldUp]) {
            case left:
                glTranslatef(-0.5, 0, 0);
                glRotatef(90 * percent, 0, -1, 0);
                glTranslatef( 0.5, 0, 0);
                break;
            case down:
                glTranslatef(0, -0.5, 0);
                glRotatef(90 * percent, 1, 0, 0);
                glTranslatef(0,  0.5, 0);
                break;
            case right:
                glTranslatef( 0.5, 0, 0);
                glRotatef(90 * percent, 0, 1, 0);
                glTranslatef(-0.5, 0, 0);
                break;
            case up:
                glTranslatef(0,  0.5, 0);
                glRotatef(90 * percent, -1, 0, 0);
                glTranslatef(0, -0.5, 0);
                break;
            default:;
        }

        switch(_oldDir) {
            case left:
            case down:
            case forward:
                glTranslatef(0, 0,  0.5);
                break;
            case right:
            case up:
            case backward:
                glTranslatef(0, 0, -0.5);
                break;
        }
    }

    gluLookAt(cEye[X], cEye[Y], cEye[Z],
              cCenter[X], cCenter[Y], cCenter[Z],
              cUp[X], cUp[Y], cUp[Z]);
}

void Snake::headLocation(int location[3]) {
    memcpy(location, _joints[0], sizeof(int) * 3);
}

Snake::~Snake() {
    for (int i = 0; i < MAX_SNAKE_LENGTH; i++)
        delete [] _joints[i];
    delete [] _joints;
}
